/**
 * @file   clptest.c
 * @author David Bremner <bremner@unb.ca>
 * @date   Tue Aug  3 10:45:44 2004
 * 
 * @brief  LP relaxation of minimum traversal using Clp
 * 
 * Read dominating set ine style file, find LP relaxation of 
 * minimum traversal.
 * Main purpose is to test Clp, particularly C interface;
 */


#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <Clp_C_Interface.h>
#include "inescan.h"
#include "rc_stack.h"

void usage(char *argv[]){
  fprintf(stderr,"%s [-i infile]\n",argv[0]);
  exit(1);
}
void fatal(const char *msg){
  fprintf(stderr,msg);
  exit(1);
}

/* Call back function - just says whenever it gets Clp0005 or Coin0002 */
static void callBack(Clp_Simplex * model, int messageNumber,
		     int nDouble, const double * vDouble,
		     int nInt, const int * vInt,
		     int nString, char ** vString) 
{
  if (messageNumber==1000002) {
    /* Coin0002 */
    assert (nString==1&&nInt==3);
    printf("Name of problem is %s\n",vString[0]);
    printf("row %d col %d el %d\n",vInt[0],vInt[1],vInt[2]);
  } else if (messageNumber==5) {
    /* Clp0005 */
    int i;
    assert (nInt==4&&nDouble==3); /* they may not all print */
    for (i=0;i<3;i++)
      printf("%d %g\n",vInt[i],vDouble[i]);
  }
  
}
		     

void expect(int token,const char *msg){
  int next_token;
  if ((next_token=inelex())!=token){
    fprintf(stderr,"line %d: expected token %d got %d  %s",ineline,token, next_token,msg);
    exit(1);
  }
}

int main(int argc, char *argv[]){
  Clp_Simplex *model = Clp_newModel();
  
  int token, row_count, dimension, stack_size,top;
  int i, offset;
  int *col_start, *row_index ;
  double  *element,*row_lb, *column_ub,*objective;
  int c;
  rc_p stack;
  
  extern FILE *inein;

  int *column_count;

  while((c=getopt(argc,argv,"i:"))!=-1){
    switch(c){
    case 'i':
      inein=fopen(optarg,"r");
      assert(inein);
      break;
    default:
      usage(argv);
      exit(1);
      
    }
  }

  
  /* scan for begin */
  for (token=inelex(); token>=0 && token != INE_BEGIN; token=inelex());
  if (token<0) 
    fatal("No begin found");

  expect(INE_EOL, "extra junk on line");

  expect(INE_INT,"no row count");
  row_count=ineint;
  stack=calloc(row_count,sizeof(rc_t));

  row_lb=calloc(row_count,sizeof(double));

  expect(INE_INT,"no dimension");
  /* take out the reference point */
  dimension=ineint-1;

  column_count=calloc(dimension,sizeof(int)+1);
  col_start=calloc(dimension+1,sizeof(int));
  column_ub=calloc(dimension,sizeof(double));
  objective=calloc(dimension,sizeof(double));

  expect(INE_EOL, "extra junk on line");
  /* we must have at least one nonzero per row */
  stack_size=row_count;
  top=0;

  for (i=0; i< row_count; i++){
    int row_size,j;

    row_lb[i]=1.0;
    expect(INE_INT,"missing row size");
    row_size=ineint;
    
    
    expect(':',"expected :");
    
    for (j=0; j<row_size; j++){
      while(stack_size<=top){
	stack_size *= 2;
	stack=realloc(stack,stack_size*sizeof(rc_t));
      }

      expect(INE_INT,"index expected");
      
      stack[top].row=i;
      /* this is not so nice, since it hard codes the fact the point 1 
	 is the reference point */
      stack[top].col=ineint-2;
      stack[top].val=1.0;
      column_count[ineint-2]++;
      top++;
    }
    expect(INE_EOL, "extra junk on line");
  }



  col_start[0]=0;
  
  for (i=1; i<dimension; i++){
    col_start[i]=col_start[i-1]+column_count[i-1];
  }

  for (i=0; i<dimension; i++){
    objective[i]=1;
    column_ub[i]=1.0;
  }

  /* this is not very well documented, but it seems the total number of 
     elements should be in the last "column start"
  */
     
  col_start[dimension]=top;


  /* somewhere to store the matrix in packed column major form */
  element=calloc(top,sizeof(double));
  row_index=calloc(top,sizeof(double));
  
  for (i=top-1; i>=0; i--){
    c=stack[i].col;
    column_count[c]--;
    
    element[col_start[c]+column_count[c]]=stack[i].val;
    row_index[col_start[c]+column_count[c]]=stack[i].row;
  }


  if (0) {
    int j;
    for (j=0; j<dimension; j++){
      printf("%d : ",j);
      for (i=col_start[j]; i<col_start[j+1]; i++){
	printf(" %d ",row_index[i]);
      }
      printf("\n");
    }
  }
  
  /* minimize */
  Clp_setOptimizationDirection(model,1);
  /* currently this is just copied from ../Clp/Samples/driverC.c */
  Clp_registerCallBack(model,callBack);


  Clp_setLogLevel(model,4);
  Clp_loadProblem(model, dimension, row_count,
		  col_start,  row_index, element, 
		  NULL, column_ub, 
		  objective, 
		  row_lb, NULL);

  Clp_writeMps(model,"test.mps",0,1,1.0);
  printf("rows=%d cols=%d\n",
	 Clp_numberRows(model),
	 Clp_numberColumns(model));
  
  Clp_dual(model,0);


  {
    /* this block lifted from driverC.c */
    /*
      Now to print out solution.  The methods used return modifiable
      arrays while the alternative names return const pointers -
      which is of course much more virtuous.
      
      This version just does non-zero columns
      
    */
    
    /* Columns */
    
    int numberColumns = Clp_numberColumns(model);
    int iColumn;
    
    
    /* Alternatively getColSolution(model) */
    double * columnPrimal = Clp_primalColumnSolution(model);
    /* Alternatively getReducedCost(model) */
    double * columnDual = Clp_dualColumnSolution(model);
    /* Alternatively getColLower(model) */
    double * columnLower = Clp_columnLower(model);
    /* Alternatively getColUpper(model) */
    double * columnUpper = Clp_columnUpper(model);
    /* Alternatively getObjCoefficients(model) */
    double * columnObjective = Clp_objective(model);
    
    printf("--------------------------------------\n");
    printf("                       Primal          Dual         Lower         Upper          Cost\n");
    
    for (iColumn=0;iColumn<numberColumns;iColumn++) {
      double value;
      value = columnPrimal[iColumn];
      if (value>1.0e-8||value<-1.0e-8) {
	printf("%6d ",iColumn);
	printf(" %13g",columnPrimal[iColumn]);
	printf(" %13g",columnDual[iColumn]);
	printf(" %13g",columnLower[iColumn]);
	printf(" %13g",columnUpper[iColumn]);
	printf(" %13g",columnObjective[iColumn]);
	printf("\n");
      }
    }
    printf("--------------------------------------\n");
  }



}
